from typing import List, Optional

from descope._auth_base import AuthBase
from descope.management.common import MgmtV1


class Role(AuthBase):
    def create(
        self,
        name: str,
        description: Optional[str] = None,
        permission_names: Optional[List[str]] = None,
        tenant_id: Optional[str] = None,
        default: Optional[bool] = None,
    ):
        """
        Create a new role.

        Args:
        name (str): role name.
        description (str): Optional description to briefly explain what this role allows.
        permission_names (List[str]): Optional list of names of permissions this role grants.
        tenant_id (str): Optional tenant ID to create the role in.
        default (bool): Optional marks this role as default role.

        Raise:
        AuthException: raised if creation operation fails
        """
        permission_names = [] if permission_names is None else permission_names

        self._auth.do_post(
            MgmtV1.role_create_path,
            {
                "name": name,
                "description": description,
                "permissionNames": permission_names,
                "tenantId": tenant_id,
                "default": default,
            },
            pswd=self._auth.management_key,
        )

    def update(
        self,
        name: str,
        new_name: str,
        description: Optional[str] = None,
        permission_names: Optional[List[str]] = None,
        tenant_id: Optional[str] = None,
        default: Optional[bool] = None,
    ):
        """
        Update an existing role with the given various fields. IMPORTANT: All parameters are used as overrides
        to the existing role. Empty fields will override populated fields. Use carefully.

        Args:
        name (str): role name.
        new_name (str): role updated name.
        description (str): Optional description to briefly explain what this role allows.
        permission_names (List[str]): Optional list of names of permissions this role grants.
        tenant_id (str): Optional tenant ID to update the role in.
        default (bool): Optional marks this role as default role.

        Raise:
        AuthException: raised if update operation fails
        """
        permission_names = [] if permission_names is None else permission_names
        self._auth.do_post(
            MgmtV1.role_update_path,
            {
                "name": name,
                "newName": new_name,
                "description": description,
                "permissionNames": permission_names,
                "tenantId": tenant_id,
                "default": default,
            },
            pswd=self._auth.management_key,
        )

    def delete(
        self,
        name: str,
        tenant_id: Optional[str] = None,
    ):
        """
        Delete an existing role. IMPORTANT: This action is irreversible. Use carefully.

        Args:
        name (str): The name of the role to be deleted.

        Raise:
        AuthException: raised if creation operation fails
        """
        self._auth.do_post(
            MgmtV1.role_delete_path,
            {"name": name, "tenantId": tenant_id},
            pswd=self._auth.management_key,
        )

    def load_all(
        self,
    ) -> dict:
        """
        Load all roles.

        Return value (dict):
        Return dict in the format
             {"roles": [{"name": <name>, "description": <description>, "permissionNames":[]}] }
        Containing the loaded role information.

        Raise:
        AuthException: raised if load operation fails
        """
        response = self._auth.do_get(
            uri=MgmtV1.role_load_all_path,
            pswd=self._auth.management_key,
        )
        return response.json()

    def search(
        self,
        tenant_ids: Optional[List[str]] = None,
        role_names: Optional[List[str]] = None,
        role_name_like: Optional[str] = None,
        permission_names: Optional[List[str]] = None,
        include_project_roles: Optional[bool] = None,
    ) -> dict:
        """
        Search roles based on the given filters.

        Args:
        tenant_ids (List[str]): List of tenant ids to filter by
        role_names (List[str]): Only return matching roles to the given names
        role_name_like (str): Return roles that contain the given string ignoring case
        permission_names (List[str]): Only return roles that have the given permissions

        Return value (dict):
        Return dict in the format
             {"roles": [{"name": <name>, "description": <description>, "permissionNames":[]}] }
        Containing the loaded role information.

        Raise:
        AuthException: raised if load operation fails
        """
        body: dict[str, str | bool | List[str]] = {}
        if tenant_ids is not None:
            body["tenantIds"] = tenant_ids
        if role_names is not None:
            body["roleNames"] = role_names
        if role_name_like is not None:
            body["roleNameLike"] = role_name_like
        if permission_names is not None:
            body["permissionNames"] = permission_names
        if include_project_roles is not None:
            body["includeProjectRoles"] = include_project_roles

        response = self._auth.do_post(
            MgmtV1.role_search_path,
            body,
            pswd=self._auth.management_key,
        )
        return response.json()



from descope.auth import Auth
from descope.management.access_key import AccessKey  # noqa: F401
from descope.management.audit import Audit  # noqa: F401
from descope.management.authz import Authz  # noqa: F401
from descope.management.fga import FGA  # noqa: F401
from descope.management.flow import Flow  # noqa: F401
from descope.management.group import Group  # noqa: F401
from descope.management.jwt import JWT  # noqa: F401
from descope.management.permission import Permission  # noqa: F401
from descope.management.project import Project  # noqa: F401
from descope.management.role import Role  # noqa: F401
from descope.management.sso_application import SSOApplication  # noqa: F401
from descope.management.sso_settings import SSOSettings  # noqa: F401
from descope.management.tenant import Tenant  # noqa: F401
from descope.management.user import User  # noqa: F401


class MGMT:
    _auth: Auth

    def __init__(self, auth: Auth):
        self._auth = auth
        self._tenant = Tenant(auth)
        self._sso_application = SSOApplication(auth)
        self._user = User(auth)
        self._access_key = AccessKey(auth)
        self._sso = SSOSettings(auth)
        self._jwt = JWT(auth)
        self._permission = Permission(auth)
        self._role = Role(auth)
        self._group = Group(auth)
        self._flow = Flow(auth)
        self._audit = Audit(auth)
        self._authz = Authz(auth)
        self._fga = FGA(auth)
        self._project = Project(auth)

    @property
    def tenant(self):
        return self._tenant

    @property
    def sso_application(self):
        return self._sso_application

    @property
    def user(self):
        return self._user

    @property
    def access_key(self):
        return self._access_key

    @property
    def sso(self):
        return self._sso

    @property
    def jwt(self):
        return self._jwt

    @property
    def permission(self):
        return self._permission

    @property
    def role(self):
        return self._role

    @property
    def group(self):
        return self._group

    @property
    def flow(self):
        return self._flow

    @property
    def audit(self):
        return self._audit

    @property
    def authz(self):
        return self._authz

    @property
    def fga(self):
        return self._fga

    @property
    def project(self):
        return self._project



from typing import Optional

from descope._auth_base import AuthBase
from descope.common import EndpointsV1, LoginOptions, validate_refresh_token_provided
from descope.exceptions import ERROR_TYPE_INVALID_ARGUMENT, AuthException


# This class is DEPRECATED please use SSO instead
class SAML(AuthBase):
    def start(
        self,
        tenant: str,
        return_url: Optional[str] = None,
        login_options: Optional[LoginOptions] = None,
        refresh_token: Optional[str] = None,
    ) -> dict:
        """
        DEPRECATED
        """
        if not tenant:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "Tenant cannot be empty"
            )

        if not return_url:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "Return url cannot be empty"
            )

        validate_refresh_token_provided(login_options, refresh_token)

        uri = EndpointsV1.auth_saml_start_path
        params = SAML._compose_start_params(tenant, return_url)
        response = self._auth.do_post(
            uri, login_options.__dict__ if login_options else {}, params, refresh_token
        )

        return response.json()

    def exchange_token(self, code: str) -> dict:
        uri = EndpointsV1.saml_exchange_token_path
        return self._auth.exchange_token(uri, code)

    @staticmethod
    def _compose_start_params(tenant: str, return_url: str) -> dict:
        res = {"tenant": tenant}
        if return_url is not None and return_url != "":
            res["redirectURL"] = return_url
        return res



from typing import Optional, Union


class UserPasswordBcrypt:
    def __init__(
        self,
        hash: str,
    ):
        """
        The bcrypt hash in plaintext format, for example "$2a$..."
        """
        self.hash = hash

    def to_dict(self) -> dict:
        return {
            "bcrypt": {
                "hash": self.hash,
            },
        }


class UserPasswordFirebase:
    def __init__(
        self,
        hash: str,
        salt: str,
        salt_separator: str,
        signer_key: str,
        memory: int,
        rounds: int,
    ):
        """
        The hash, salt, salt separator, and signer key should be base64 strings using
        standard encoding with padding.
        The memory cost value is an integer, usually between 12 to 17.
        The rounds cost value is an integer, usually between 6 to 10.
        """
        self.hash = hash
        self.salt = salt
        self.salt_separator = salt_separator
        self.signer_key = signer_key
        self.memory = memory
        self.rounds = rounds

    def to_dict(self) -> dict:
        return {
            "firebase": {
                "hash": self.hash,
                "salt": self.salt,
                "saltSeparator": self.salt_separator,
                "signerKey": self.signer_key,
                "memory": self.memory,
                "rounds": self.rounds,
            },
        }


class UserPasswordPbkdf2:
    def __init__(
        self,
        hash: str,
        salt: str,
        iterations: int,
        variant: str,
    ):
        """
        The hash and salt should be base64 strings using standard encoding with padding.
        The iterations cost value is an integer, usually in the thousands.
        The hash variant should be either "sha1", "sha256", or "sha512".
        """
        self.hash = hash
        self.salt = salt
        self.iterations = iterations
        self.variant = variant

    def to_dict(self) -> dict:
        return {
            "pbkdf2": {
                "hash": self.hash,
                "salt": self.salt,
                "iterations": self.iterations,
                "type": self.variant,
            },
        }


class UserPasswordDjango:
    def __init__(
        self,
        hash: str,
    ):
        """
        The django hash in plaintext format, for example "pbkdf2_sha256$..."
        """
        self.hash = hash

    def to_dict(self) -> dict:
        return {
            "django": {
                "hash": self.hash,
            },
        }


class UserPassword:
    def __init__(
        self,
        cleartext: Optional[str] = None,
        hashed: Optional[
            Union[
                UserPasswordBcrypt,
                UserPasswordFirebase,
                UserPasswordPbkdf2,
                UserPasswordDjango,
            ]
        ] = None,
    ):
        """
        Set a UserPassword on UserObj objects when calling invite_batch to create or invite users
        with a cleartext or prehashed password. Note that only one of the two options should be set.
        """
        self.cleartext = cleartext
        self.hashed = hashed



from datetime import datetime, timezone
from typing import Any, List, Optional

from descope._auth_base import AuthBase
from descope.management.common import MgmtV1


class Authz(AuthBase):
    def save_schema(self, schema: dict, upgrade: bool = False):
        """
        Create or update the ReBAC schema.
        In case of update, will update only given namespaces and will not delete namespaces unless upgrade flag is true.
        Args:
        schema (dict): the schema dict with format
            {
                "name": "name-of-schema",
                "namespaces": [
                    {
                        "name": "name-of-namespace",
                        "relationDefinitions": [
                            {
                                "name": "name-of-relation-definition",
                                "complexDefinition": {
                                    "nType": "one of child|union|intersect|sub",
                                    "children": "optional list of node children - same format as complexDefinition",
                                    "expression": {
                                        "neType": "one of self|targetSet|relationLeft|relationRight",
                                        "relationDefinition": "name of relation definition for relationLeft and relationRight",
                                        "relationDefinitionNamespace": "the namespace for the rd above",
                                        "targetRelationDefinition": "relation definition for targetSet and relationLeft/right",
                                        "targetRelationDefinitionNamespace": "the namespace for above"
                                    }
                                }
                            }
                        ]
                    }
                ]
            }
        Schema name can be used for projects to track versioning.
        Raise:
        AuthException: raised if saving fails
        """
        self._auth.do_post(
            MgmtV1.authz_schema_save,
            {"schema": schema, "upgrade": upgrade},
            pswd=self._auth.management_key,
        )

    def delete_schema(self):
        """
        Delete the schema for the project which will also delete all relations.
        Raise:
        AuthException: raised if delete schema fails
        """
        self._auth.do_post(
            MgmtV1.authz_schema_delete,
            None,
            pswd=self._auth.management_key,
        )

    def load_schema(self) -> dict:
        """
        Load the schema for the project
        Return value (dict):
        Return dict in the format of schema as above (see save_schema)
        Raise:
        AuthException: raised if load schema fails
        """
        response = self._auth.do_post(
            MgmtV1.authz_schema_load,
            None,
            pswd=self._auth.management_key,
        )
        return response.json()["schema"]

    def save_namespace(
        self, namespace: dict, old_name: str = "", schema_name: str = ""
    ):
        """
        Create or update the given namespace
        Will not delete relation definitions not mentioned in the namespace.
        Args:
        namespace (dict): namespace in the format as specified above (see save_schema)
        old_name (str): is used if we are changing the namespace name
        schema_name (str): is optional and can be used to track the current schema version.
        Raise:
        AuthException: raised if save namespace fails
        """
        body: dict[str, Any] = {"namespace": namespace}
        if old_name != "":
            body["oldName"] = old_name
        if schema_name != "":
            body["schemaName"] = schema_name
        self._auth.do_post(
            MgmtV1.authz_ns_save,
            body,
            pswd=self._auth.management_key,
        )

    def delete_namespace(self, name: str, schema_name: str = ""):
        """
        delete_namespace will also delete the relevant relations.
        Args:
        name (str): namespace name to delete
        schema_name (str): is optional and can be used to track the current schema version.
        Raise:
        AuthException: raised if delete namespace fails
        """
        body: dict[str, Any] = {"name": name}
        if schema_name != "":
            body["schemaName"] = schema_name
        self._auth.do_post(
            MgmtV1.authz_ns_delete,
            body,
            pswd=self._auth.management_key,
        )

    def save_relation_definition(
        self,
        relation_definition: dict,
        namespace: str,
        old_name: str = "",
        schema_name: str = "",
    ):
        """
        Create or update the given relation definition
        Will not delete relation definitions not mentioned in the namespace.
        Args:
        relation_definition (dict): relation definition in the format as specified above (see save_schema)
        namespace (str): the namespace for the relation definition
        old_name (str): is used if we are changing the relation definition name
        schema_name (str): is optional and can be used to track the current schema version.
        Raise:
        AuthException: raised if save relation definition fails
        """
        body: dict[str, Any] = {
            "relationDefinition": relation_definition,
            "namespace": namespace,
        }
        if old_name != "":
            body["oldName"] = old_name
        if schema_name != "":
            body["schemaName"] = schema_name
        self._auth.do_post(
            MgmtV1.authz_rd_save,
            body,
            pswd=self._auth.management_key,
        )

    def delete_relation_definition(
        self, name: str, namespace: str, schema_name: str = ""
    ):
        """
        delete_relation_definition will also delete the relevant relations.
        Args:
        name (str): relation definition name to delete
        namespace (str): the namespace for the relation definition
        schema_name (str): is optional and can be used to track the current schema version.
        Raise:
        AuthException: raised if delete namespace fails
        """
        body: dict[str, Any] = {"name": name, "namespace": namespace}
        if schema_name != "":
            body["schemaName"] = schema_name
        self._auth.do_post(
            MgmtV1.authz_rd_delete,
            body,
            pswd=self._auth.management_key,
        )

    def create_relations(
        self,
        relations: List[dict],
    ):
        """
        Create the given relations based on the existing schema
        Args:
        relations (List[dict]): the relations to create. Each in the following format:
            {
                "resource": "id of the resource that has the relation",
                "relationDefinition": "the relation definition for the relation",
                "namespace": "namespace for the relation definition",
                "target": "the target that has the relation - usually users or other resources",
                "targetSetResource": "if the target is a group that has another relation",
                "targetSetRelationDefinition": "the relation definition for the targetSet group",
                "targetSetRelationDefinitionNamespace": "the namespace for the relation definition for the targetSet group",
                "query": {
                    "tenants": ["t1", "t2"],
                    "roles": ["r1", "r2"],
                    "text": "full-text-search",
                    "statuses": ["enabled|disabled|invited"],
                    "ssoOnly": True|False,
                    "withTestUser": True|False,
                    "customAttributes": {
                        "key": "value",
                        ...
                    }
                }
            }
            Each relation should have exactly one of: target, targetSet, query
            Regarding query above, it should be specified if the target is a set of users that matches the query - all fields are optional
        Raise:
        AuthException: raised if create relations fails
        """
        self._auth.do_post(
            MgmtV1.authz_re_create,
            {
                "relations": relations,
            },
            pswd=self._auth.management_key,
        )

    def delete_relations(
        self,
        relations: List[dict],
    ):
        """
        Delete the given relations based on the existing schema
        Args:
        relations (List[dict]): the relations to create. Each in the format as specified above for (create_relations)
        Raise:
        AuthException: raised if delete relations fails
        """
        self._auth.do_post(
            MgmtV1.authz_re_delete,
            {
                "relations": relations,
            },
            pswd=self._auth.management_key,
        )

    def delete_relations_for_resources(
        self,
        resources: List[str],
    ):
        """
        Delete all relations to the given resources
        Args:
        resources (List[str]): the list of resources to delete any relations for
        Raise:
        AuthException: raised if delete relations for resources fails
        """
        self._auth.do_post(
            MgmtV1.authz_re_delete_resources,
            {
                "resources": resources,
            },
            pswd=self._auth.management_key,
        )

    def has_relations(
        self,
        relation_queries: List[dict],
    ) -> List[dict]:
        """
        Queries the given relations to see if they exist returning true if they do
        Args:
        relation_queries (List[dict]): List of queries each in the format of:
            {
                "resource": "resource for the relation query",
                "relationDefinition": "the relation definition for the relation query",
                "namespace": "namespace for the relation definition",
                "target": "the target that has the relation - usually users or other resources"
            }

        Return value (List[dict]):
        Return List in the format
             [
                {
                    "resource": "resource for the relation query",
                    "relationDefinition": "the relation definition for the relation query",
                    "namespace": "namespace for the relation definition",
                    "target": "the target that has the relation - usually users or other resources",
                    "hasRelation": True|False
                }
            ]
        Raise:
        AuthException: raised if query fails
        """
        response = self._auth.do_post(
            MgmtV1.authz_re_has_relations,
            {
                "relationQueries": relation_queries,
            },
            pswd=self._auth.management_key,
        )
        return response.json()["relationQueries"]

    def who_can_access(
        self, resource: str, relation_definition: str, namespace: str
    ) -> List[dict]:
        """
        Finds the list of targets (usually users) who can access the given resource with the given RD
        Args:
        resource (str): the resource we are checking
        relation_definition (str): the RD we are checking
        namespace (str): the namespace for the RD

        Return value (List[str]): list of targets (user IDs usually that have the access)
        Raise:
        AuthException: raised if query fails
        """
        response = self._auth.do_post(
            MgmtV1.authz_re_who,
            {
                "resource": resource,
                "relationDefinition": relation_definition,
                "namespace": namespace,
            },
            pswd=self._auth.management_key,
        )
        return response.json()["targets"]

    def resource_relations(self, resource: str) -> List[dict]:
        """
        Returns the list of all defined relations (not recursive) on the given resource.
        Args:
        resource (str): the resource we are listing relations for

        Return value (List[dict]):
        Return List of relations each in the format of a relation as documented in create_relations
        Raise:
        AuthException: raised if query fails
        """
        response = self._auth.do_post(
            MgmtV1.authz_re_resource,
            {"resource": resource},
            pswd=self._auth.management_key,
        )
        return response.json()["relations"]

    def targets_relations(self, targets: List[str]) -> List[dict]:
        """
        Returns the list of all defined relations (not recursive) for the given targets.
        Args:
        targets (List[str]): the list of targets we are returning the relations for

        Return value (List[dict]):
        Return List of relations each in the format of a relation as documented in create_relations
        Raise:
        AuthException: raised if query fails
        """
        response = self._auth.do_post(
            MgmtV1.authz_re_targets,
            {"targets": targets},
            pswd=self._auth.management_key,
        )
        return response.json()["relations"]

    def what_can_target_access(self, target: str) -> List[dict]:
        """
        Returns the list of all relations for the given target including derived relations from the schema tree.
        Args:
        target (str): the target we are returning the relations for

        Return value (List[dict]):
        Return List of relations each in the format of a relation as documented in create_relations
        Raise:
        AuthException: raised if query fails
        """
        response = self._auth.do_post(
            MgmtV1.authz_re_target_all,
            {"target": target},
            pswd=self._auth.management_key,
        )
        return response.json()["relations"]

    def what_can_target_access_with_relation(
        self, target: str, relation_definition: str, namespace: str
    ) -> List[dict]:
        """
        Returns the list of all resources that the target has the given relation to including all derived relations
        Args:
        target (str): the target we are returning the relations for
        relation_definition (str): the RD we are checking
        namespace (str): the namespace for the RD

        Return value (List[dict]):
        Return List of relations each in the format of a relation as documented in create_relations
        Raise:
        AuthException: raised if query fails
        """
        response = self._auth.do_post(
            MgmtV1.authz_re_target_with_relation,
            {
                "target": target,
                "relationDefinition": relation_definition,
                "namespace": namespace,
            },
            pswd=self._auth.management_key,
        )
        return response.json()["relations"]

    def get_modified(self, since: Optional[datetime] = None) -> dict:
        """
        Get all targets and resources changed since the given date.
        Args:
        since (datetime): only return changes from this given datetime

        Return value (dict):
        Dict including "resources" list of strings, "targets" list of strings and "schemaChanged" bool
        Raise:
        AuthException: raised if query fails
        """
        response = self._auth.do_post(
            MgmtV1.authz_get_modified,
            {
                "since": (
                    int(since.replace(tzinfo=timezone.utc).timestamp() * 1000)
                    if since
                    else 0
                )
            },
            pswd=self._auth.management_key,
        )
        return response.json()["relations"]



from __future__ import annotations

from typing import Iterable

import requests

from descope.auth import Auth  # noqa: F401
from descope.authmethod.enchantedlink import EnchantedLink  # noqa: F401
from descope.authmethod.magiclink import MagicLink  # noqa: F401
from descope.authmethod.oauth import OAuth  # noqa: F401
from descope.authmethod.otp import OTP  # noqa: F401
from descope.authmethod.password import Password  # noqa: F401
from descope.authmethod.saml import SAML  # noqa: F401
from descope.authmethod.sso import SSO  # noqa: F401
from descope.authmethod.totp import TOTP  # noqa: F401
from descope.authmethod.webauthn import WebAuthn  # noqa: F401
from descope.common import DEFAULT_TIMEOUT_SECONDS, AccessKeyLoginOptions, EndpointsV1
from descope.exceptions import ERROR_TYPE_INVALID_ARGUMENT, AuthException
from descope.mgmt import MGMT  # noqa: F401


class DescopeClient:
    ALGORITHM_KEY = "alg"

    def __init__(
        self,
        project_id: str,
        public_key: dict | None = None,
        skip_verify: bool = False,
        management_key: str | None = None,
        timeout_seconds: float = DEFAULT_TIMEOUT_SECONDS,
        jwt_validation_leeway: int = 5,
    ):
        auth = Auth(
            project_id,
            public_key,
            skip_verify,
            management_key,
            timeout_seconds,
            jwt_validation_leeway,
        )
        self._auth = auth
        self._mgmt = MGMT(auth)
        self._magiclink = MagicLink(auth)
        self._enchantedlink = EnchantedLink(auth)
        self._oauth = OAuth(auth)
        self._saml = SAML(auth)  # deprecated
        self._sso = SSO(auth)
        self._otp = OTP(auth)
        self._totp = TOTP(auth)
        self._webauthn = WebAuthn(auth)
        self._password = Password(auth)

    @property
    def mgmt(self):
        if not self._auth.management_key:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "management_key cannot be empty"
            )
        return self._mgmt

    @property
    def magiclink(self):
        return self._magiclink

    @property
    def enchantedlink(self):
        return self._enchantedlink

    @property
    def otp(self):
        return self._otp

    @property
    def totp(self):
        return self._totp

    @property
    def oauth(self):
        return self._oauth

    # ######## deprecated (use sso instead)
    @property
    def saml(self):
        return self._saml

    # ###################

    @property
    def sso(self):
        return self._sso

    @property
    def webauthn(self):
        return self._webauthn

    @property
    def password(self):
        return self._password

    def validate_permissions(self, jwt_response: dict, permissions: list[str]) -> bool:
        """
        Validate that a jwt_response has been granted the specified permissions.
            For a multi-tenant environment use validate_tenant_permissions function

        Args:
        jwt_response (dict): The jwt_response object which includes all JWT claims information
        permissions (List[str]): List of permissions to validate for this jwt_response

        Return value (bool): returns true if all permissions granted; false if at least one permission not granted
        """
        return self.validate_tenant_permissions(jwt_response, "", permissions)

    def get_matched_permissions(
        self, jwt_response: dict, permissions: list[str]
    ) -> list[str]:
        """
        Get the list of permissions that a jwt_response has been granted from the provided list of permissions.
            For a multi-tenant environment use get_matched_tenant_permissions function

        Args:
        jwt_response (dict): The jwt_response object which includes all JWT claims information
        permissions (List[str]): List of permissions to validate for this jwt_response

        Return value (List[str]): returns the list of permissions that are granted
        """
        return self.get_matched_tenant_permissions(jwt_response, "", permissions)

    def validate_tenant_permissions(
        self, jwt_response: dict, tenant: str, permissions: list[str]
    ) -> bool:
        """
        Validate that a jwt_response has been granted the specified permissions on the specified tenant.
            For a multi-tenant environment use validate_tenant_permissions function

        Args:
        jwt_response (dict): The jwt_response object which includes all JWT claims information
        tenant (str): TenantId
        permissions (List[str]): List of permissions to validate for this jwt_response

        Return value (bool): returns true if all permissions granted; false if at least one permission not granted
        """
        if not jwt_response:
            return False

        if isinstance(permissions, str):
            permissions = [permissions]

        granted = []
        if tenant == "":
            granted = jwt_response.get("permissions", [])
        else:
            # ensure that the tenant is associated with the jwt_response
            if tenant not in jwt_response.get("tenants", {}):
                return False
            granted = (
                jwt_response.get("tenants", {}).get(tenant, {}).get("permissions", [])
            )

        for perm in permissions:
            if perm not in granted:
                return False
        return True

    def get_matched_tenant_permissions(
        self, jwt_response: dict, tenant: str, permissions: list[str]
    ) -> list[str]:
        """
        Get the list of permissions that a jwt_response has been granted from the provided list of permissions on the specified tenant.
            For a multi-tenant environment use get_matched_tenant_permissions function

        Args:
        jwt_response (dict): The jwt_response object which includes all JWT claims information
        tenant (str): TenantId
        permissions (List[str]): List of permissions to validate for this jwt_response

        Return value (List[str]): returns the list of permissions that are granted
        """
        if not jwt_response:
            return []

        if isinstance(permissions, str):
            permissions = [permissions]

        granted = []
        if tenant == "":
            granted = jwt_response.get("permissions", [])
        else:
            # ensure that the tenant is associated with the jwt_response
            if tenant not in jwt_response.get("tenants", {}):
                return []
            granted = (
                jwt_response.get("tenants", {}).get(tenant, {}).get("permissions", [])
            )

        matched = []
        for perm in permissions:
            if perm in granted:
                matched.append(perm)
        return matched

    def validate_roles(self, jwt_response: dict, roles: list[str]) -> bool:
        """
        Validate that a jwt_response has been granted the specified roles.
            For a multi-tenant environment use validate_tenant_roles function

        Args:
        jwt_response (dict): The jwt_response object which includes all JWT claims information
        roles (List[str]): List of roles to validate for this jwt_response

        Return value (bool): returns true if all roles granted; false if at least one role not granted
        """
        return self.validate_tenant_roles(jwt_response, "", roles)

    def get_matched_roles(self, jwt_response: dict, roles: list[str]) -> list[str]:
        """
        Get the list of roles that a jwt_response has been granted from the provided list of roles.
            For a multi-tenant environment use get_matched_tenant_roles function

        Args:
        jwt_response (dict): The jwt_response object which includes all JWT claims information
        roles (List[str]): List of roles to validate for this jwt_response

        Return value (List[str]): returns the list of roles that are granted
        """
        return self.get_matched_tenant_roles(jwt_response, "", roles)

    def validate_tenant_roles(
        self, jwt_response: dict, tenant: str, roles: list[str]
    ) -> bool:
        """
        Validate that a jwt_response has been granted the specified roles on the specified tenant.
            For a multi-tenant environment use validate_tenant_roles function

        Args:
        jwt_response (dict): The jwt_response object which includes all JWT claims information
        tenant (str): TenantId
        roles (List[str]): List of roles to validate for this jwt_response

        Return value (bool): returns true if all roles granted; false if at least one role not granted
        """
        if not jwt_response:
            return False

        if isinstance(roles, str):
            roles = [roles]

        granted = []
        if tenant == "":
            granted = jwt_response.get("roles", [])
        else:
            # ensure that the tenant is associated with the jwt_response
            if tenant not in jwt_response.get("tenants", {}):
                return False
            granted = jwt_response.get("tenants", {}).get(tenant, {}).get("roles", [])

        for role in roles:
            if role not in granted:
                return False
        return True

    def get_matched_tenant_roles(
        self, jwt_response: dict, tenant: str, roles: list[str]
    ) -> list[str]:
        """
        Get the list of roles that a jwt_response has been granted from the provided list of roles on the specified tenant.
            For a multi-tenant environment use get_matched_tenant_roles function

        Args:
        jwt_response (dict): The jwt_response object which includes all JWT claims information
        tenant (str): TenantId
        roles (List[str]): List of roles to validate for this jwt_response

        Return value (List[str]): returns the list of roles that are granted
        """
        if not jwt_response:
            return []

        if isinstance(roles, str):
            roles = [roles]

        granted = []
        if tenant == "":
            granted = jwt_response.get("roles", [])
        else:
            # ensure that the tenant is associated with the jwt_response
            if tenant not in jwt_response.get("tenants", {}):
                return []
            granted = jwt_response.get("tenants", {}).get(tenant, {}).get("roles", [])

        matched = []
        for role in roles:
            if role in granted:
                matched.append(role)
        return matched

    def validate_session(
        self, session_token: str, audience: str | Iterable[str] | None = None
    ) -> dict:
        """
        Validate a session token. Call this function for every incoming request to your
        private endpoints. Alternatively, use validate_and_refresh_session in order to
        automatically refresh expired sessions. If you need to use these specific claims
        [amr, drn, exp, iss, rexp, sub, jwt] in the top level of the response dict, please use
        them from the sessionToken key instead, as these claims will soon be deprecated from the top level
        of the response dict.

        Args:
        session_token (str): The session token to be validated
        audience (str|Iterable[str]|None): Optional recipients that the JWT is intended for (must be equal to the 'aud' claim on the provided token)

        Return value (dict):
        Return dict includes the session token and all JWT claims

        Raise:
        AuthException: Exception is raised if session is not authorized or any other error occurs
        """
        return self._auth.validate_session(session_token, audience)

    def refresh_session(
        self, refresh_token: str, audience: str | Iterable[str] | None = None
    ) -> dict:
        """
        Refresh a session. Call this function when a session expires and needs to be refreshed.

        Args:
        refresh_token (str): The refresh token that will be used to refresh the session
        audience (str|Iterable[str]|None): Optional recipients that the JWT is intended for (must be equal to the 'aud' claim on the provided token)

        Return value (dict):
        Return dict includes the session token, refresh token, and all JWT claims

        Raise:
        AuthException: Exception is raised if refresh token is not authorized or any other error occurs
        """
        return self._auth.refresh_session(refresh_token, audience)

    def validate_and_refresh_session(
        self,
        session_token: str,
        refresh_token: str,
        audience: str | Iterable[str] | None = None,
    ) -> dict:
        """
        Validate the session token and refresh it if it has expired, the session token will automatically be refreshed.
        Either the session_token or the refresh_token must be provided.
        Call this function for every incoming request to your
        private endpoints. Alternatively, use validate_session to only validate the session.

        Args:
        session_token (str): The session token to be validated
        refresh_token (str): The refresh token that will be used to refresh the session token, if needed
        audience (str|Iterable[str]|None): Optional recipients that the JWT is intended for (must be equal to the 'aud' claim on the provided token)

        Return value (dict):
        Return dict includes the session token, refresh token, and all JWT claims

        Raise:
        AuthException: Exception is raised if session is not authorized or another error occurs
        """
        return self._auth.validate_and_refresh_session(
            session_token, refresh_token, audience
        )

    def logout(self, refresh_token: str) -> requests.Response:
        """
        Logout user from current session and revoke the refresh_token. After calling this function,
            you must invalidate or remove any cookies you have created.

        Args:
        refresh_token (str): The refresh token

        Return value (requests.Response): returns the response from the Descope server

        Raise:
        AuthException: Exception is raised if session is not authorized or another error occurs
        """
        if refresh_token is None:
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_ARGUMENT,
                f"signed refresh token {refresh_token} is empty",
            )

        uri = EndpointsV1.logout_path
        return self._auth.do_post(uri, {}, None, refresh_token)

    def logout_all(self, refresh_token: str) -> requests.Response:
        """
        Logout user from all active sessions and revoke the refresh_token. After calling this function,
            you must invalidate or remove any cookies you have created.

        Args:
        refresh_token (str): The refresh token

        Return value (requests.Response): returns the response from the Descope server

        Raise:
        AuthException: Exception is raised if session is not authorized or another error occurs
        """
        if refresh_token is None:
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_ARGUMENT,
                f"signed refresh token {refresh_token} is empty",
            )

        uri = EndpointsV1.logout_all_path
        return self._auth.do_post(uri, {}, None, refresh_token)

    def me(self, refresh_token: str) -> dict:
        """
        Retrieve user details for the refresh token. The returned data includes email, name, phone,
            list of loginIds and boolean flags for verifiedEmail, verifiedPhone.

        Args:
        refresh_token (str): The refresh token

        Return value (dict): returns the user details from the server
            (email:str, name:str, phone:str, loginIds[str], verifiedEmail:bool, verifiedPhone:bool)

        Raise:
        AuthException: Exception is raised if session is not authorized or another error occurs
        """
        if refresh_token is None:
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_ARGUMENT,
                f"signed refresh token {refresh_token} is empty",
            )

        uri = EndpointsV1.me_path
        response = self._auth.do_get(
            uri=uri, params=None, allow_redirects=None, pswd=refresh_token
        )
        return response.json()

    def my_tenants(
        self,
        refresh_token: str,
        dct: bool = False,
        ids: list[str] | None = None,
    ) -> dict:
        """
        Retrieve tenant attributes that user belongs to, one of dct/ids must be populated .

        Args:
        dct (bool): Get only the selected tenant from jwt
        ids (List[str]): Get the list of tenants
        refresh_token (str): The refresh token

        Return value (dict): returns the tenant requested from the server
            (id:str, name:str, customAttributes:dict)

        Raise:
        AuthException: Exception is raised if session is not authorized or another error occurs
        """
        if refresh_token is None:
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_ARGUMENT,
                f"signed refresh token {refresh_token} is empty",
            )
        if dct is True and ids is not None and len(ids) > 0:
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_ARGUMENT,
                "Only one of 'dct' or 'ids' should be supplied",
            )
        if dct is False and (ids is None or len(ids) == 0):
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_ARGUMENT,
                "Only one of 'dct' or 'ids' should be supplied",
            )

        body: dict[str, bool | list[str]] = {"dct": dct}
        if ids is not None:
            body["ids"] = ids

        uri = EndpointsV1.my_tenants_path
        response = self._auth.do_post(uri, body, None, refresh_token)
        return response.json()

    def history(self, refresh_token: str) -> list[dict]:
        """
        Retrieve user authentication history for the refresh token

        Args:
        refresh_token (str): The refresh token

        Return value (List[dict]):
        Return List in the format
             [
                {
                    "userId": "User's ID",
                    "loginTime": "User'sLogin time",
                    "city": "User's city",
                    "country": "User's country",
                    "ip": User's IP
                }
            ]

        Raise:
        AuthException: Exception is raised if session is not authorized or another error occurs
        """
        if refresh_token is None:
            raise AuthException(
                400,
                ERROR_TYPE_INVALID_ARGUMENT,
                f"signed refresh token {refresh_token} is empty",
            )

        uri = EndpointsV1.history_path
        response = self._auth.do_get(
            uri=uri, params=None, allow_redirects=None, pswd=refresh_token
        )
        return response.json()

    def exchange_access_key(
        self,
        access_key: str,
        audience: str | Iterable[str] | None = None,
        login_options: AccessKeyLoginOptions | None = None,
    ) -> dict:
        """
        Return a new session token for the given access key

        Args:
        access_key (str): The access key
        audience (str|Iterable[str]|None): Optional recipients that the JWT is intended for (must be equal to the 'aud' claim on the provided token)
        login_options (AccessKeyLoginOptions): Optional advanced controls over login parameters

        Return dict includes the session token and all JWT claims

        Raise:
        AuthException: Exception is raised if access key is not valid or another error occurs
        """
        if not access_key:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "Access key cannot be empty"
            )

        return self._auth.exchange_access_key(access_key, audience, login_options)

    def select_tenant(
        self,
        tenant_id: str,
        refresh_token: str,
    ) -> dict:
        """
        Add to JWT a selected tenant claim

        Args:
        refresh_token (str): The refresh token that will be used to refresh the session token, if needed
        tenant_id (str): The tenant id to place on JWT

        Return value (dict):
        Return dict includes the session token, refresh token, with the tenant id on the jwt

        Raise:
        AuthException: Exception is raised if session is not authorized or another error occurs
        """
        return self._auth.select_tenant(tenant_id, refresh_token)



from datetime import datetime
from typing import Any, List, Optional

from descope._auth_base import AuthBase
from descope.management.common import MgmtV1


class Audit(AuthBase):
    def search(
        self,
        user_ids: Optional[List[str]] = None,
        actions: Optional[List[str]] = None,
        excluded_actions: Optional[List[str]] = None,
        devices: Optional[List[str]] = None,
        methods: Optional[List[str]] = None,
        geos: Optional[List[str]] = None,
        remote_addresses: Optional[List[str]] = None,
        login_ids: Optional[List[str]] = None,
        tenants: Optional[List[str]] = None,
        no_tenants: bool = False,
        text: Optional[str] = None,
        from_ts: Optional[datetime] = None,
        to_ts: Optional[datetime] = None,
    ) -> dict:
        """
        Search the audit trail up to last 30 days based on given parameters

        Args:
        user_ids (List[str]): Optional list of user IDs to filter by
        actions (List[str]): Optional list of actions to filter by
        excluded_actions (List[str]): Optional list of actions to exclude
        devices (List[str]): Optional list of devices to filter by. Current devices supported are "Bot"/"Mobile"/"Desktop"/"Tablet"/"Unknown"
        methods (List[str]): Optional list of methods to filter by. Current auth methods are "otp"/"totp"/"magiclink"/"oauth"/"saml"/"password"
        geos (List[str]): Optional list of geos to filter by. Geo is currently country code like "US", "IL", etc.
        remote_addresses (List[str]): Optional list of remote addresses to filter by
        login_ids (List[str]): Optional list of login IDs to filter by
        tenants (List[str]): Optional list of tenants to filter by
        no_tenants (bool): Should audits without any tenants always be included
        text (str): Free text search across all fields
        from_ts (datetime): Retrieve records newer than given time but not older than 30 days
        to_ts (datetime): Retrieve records older than given time

        Return value (dict):
        Return dict in the format
             {
                "audits": [
                    {
                        "projectId":"",
                        "userId": "",
                        "action": "",
                        "occurred": 0 (unix-time-milli),
                        "device": "",
                        "method": "",
                        "geo": "",
                        "remoteAddress": "",
                        "externalIds": [""],
                        "tenants": [""],
                        "data": {
                            "field1": "field1-value",
                            "more-details": "in-console-examples"
                        }
                    }
                ]
            }
        Raise:
        AuthException: raised if search operation fails
        """
        body: dict[str, Any] = {"noTenants": no_tenants}
        if user_ids is not None:
            body["userIds"] = user_ids
        if actions is not None:
            body["actions"] = actions
        if excluded_actions is not None:
            body["excludedActions"] = excluded_actions
        if devices is not None:
            body["devices"] = devices
        if methods is not None:
            body["methods"] = methods
        if geos is not None:
            body["geos"] = geos
        if remote_addresses is not None:
            body["remoteAddresses"] = remote_addresses
        if login_ids is not None:
            body["externalIds"] = login_ids
        if tenants is not None:
            body["tenants"] = tenants
        if text is not None:
            body["text"] = text
        if from_ts is not None:
            body["from"] = int(from_ts.timestamp() * 1000)
        if to_ts is not None:
            body["to"] = int(to_ts.timestamp() * 1000)

        response = self._auth.do_post(
            MgmtV1.audit_search,
            body=body,
            pswd=self._auth.management_key,
        )
        return {
            "audits": list(map(Audit._convert_audit_record, response.json()["audits"]))
        }

    def create_event(
        self,
        action: str,
        type: str,
        actor_id: str,
        tenant_id: str,
        user_id: Optional[str] = None,
        data: Optional[dict] = None,
    ):
        """
        Create audit event based on given parameters

        Args:
        action (str): Audit action
        type (str): Audit type (info/warn/error)
        actor_id (str): Audit actor id
        tenant_id (str): Audit tenant id
        user_id (str): Optional, Audit user id
        data (dict): Optional, Audit data

        Raise:
        AuthException: raised if search operation fails
        """
        body: dict[str, Any] = {
            "action": action,
            "type": type,
            "actorId": actor_id,
            "tenantId": tenant_id,
        }
        if user_id is not None:
            body["userId"] = user_id
        if data is not None:
            body["data"] = data

        self._auth.do_post(
            MgmtV1.audit_create_event,
            body=body,
            pswd=self._auth.management_key,
        )

    @staticmethod
    def _convert_audit_record(a: dict) -> dict:
        return {
            "projectId": a.get("projectId", ""),
            "userId": a.get("userId", ""),
            "action": a.get("action", ""),
            "occurred": datetime.utcfromtimestamp(float(a.get("occurred", "0")) / 1000),
            "device": a.get("device", ""),
            "method": a.get("method", ""),
            "geo": a.get("geo", ""),
            "remoteAddress": a.get("remoteAddress", ""),
            "loginIds": a.get("externalIds", []),
            "tenants": a.get("tenants", []),
            "data": a.get("data", {}),
        }



from typing import List, Optional

from descope._auth_base import AuthBase
from descope.management.common import (
    AssociatedTenant,
    MgmtV1,
    associated_tenants_to_dict,
)


class AccessKey(AuthBase):
    def create(
        self,
        name: str,
        expire_time: int = 0,
        role_names: Optional[List[str]] = None,
        key_tenants: Optional[List[AssociatedTenant]] = None,
        user_id: Optional[str] = None,
        custom_claims: Optional[dict] = None,
        description: Optional[str] = None,
        permitted_ips: Optional[List[str]] = None,
    ) -> dict:
        """
        Create a new access key.

        Args:
        name (str): Access key name.
        expire_time (int): Access key expiration. Leave at 0 to make it indefinite.
        role_names (List[str]): An optional list of the access key's roles without tenant association. These roles are
            mutually exclusive with the `key_tenant` roles, which take precedence over them.
        key_tenants (List[AssociatedTenant]): An optional list of the access key's tenants, and optionally, their roles per tenant. These roles are
            mutually exclusive with the general `role_names`, and take precedence over them.
        user_id (str): Bind access key to this user id
            If user_id is supplied, then authorizations will be ignored, and the access key will be bound to the user's authorization.
        custom_claims (dict): Optional, map of claims and their values that will be present in the JWT.
        description (str): an optional text the access key can hold.
        permitted_ips: (List[str]): An optional list of IP addresses or CIDR ranges that are allowed to use the access key.

        Return value (dict):
        Return dict in the format
            {
                "key": {},
                "cleartext": {}
            }
        Containing the created access key information and its cleartext. The key cleartext will only be returned on creation.
        Make sure to save it securely.

        Raise:
        AuthException: raised if create operation fails
        """
        role_names = [] if role_names is None else role_names
        key_tenants = [] if key_tenants is None else key_tenants

        response = self._auth.do_post(
            MgmtV1.access_key_create_path,
            AccessKey._compose_create_body(
                name,
                expire_time,
                role_names,
                key_tenants,
                user_id,
                custom_claims,
                description,
                permitted_ips,
            ),
            pswd=self._auth.management_key,
        )
        return response.json()

    def load(
        self,
        id: str,
    ) -> dict:
        """
        Load an existing access key.

        Args:
        id (str): The id of the access key to be loaded.

        Return value (dict):
        Return dict in the format
             {"key": {}}
        Containing the loaded access key information.

        Raise:
        AuthException: raised if load operation fails
        """
        response = self._auth.do_get(
            uri=MgmtV1.access_key_load_path,
            params={"id": id},
            pswd=self._auth.management_key,
        )
        return response.json()

    def search_all_access_keys(
        self,
        tenant_ids: Optional[List[str]] = None,
    ) -> dict:
        """
        Search all access keys.

        Args:
        tenant_ids (List[str]): Optional list of tenant IDs to filter by

        Return value (dict):
        Return dict in the format
             {"keys": []}
        "keys" contains a list of all of the found users and their information

        Raise:
        AuthException: raised if search operation fails
        """
        tenant_ids = [] if tenant_ids is None else tenant_ids

        response = self._auth.do_post(
            MgmtV1.access_keys_search_path,
            {"tenantIds": tenant_ids},
            pswd=self._auth.management_key,
        )
        return response.json()

    def update(
        self,
        id: str,
        name: str,
        description: Optional[str] = None,
    ):
        """
        Update an existing access key with the given various fields. IMPORTANT: id and name are mandatory fields.

        Args:
        id (str): The id of the access key to update.
        name (str): The updated access key name.
        description (str): The description of the access key to update. If not provided, it will not be overriden.

        Raise:
        AuthException: raised if update operation fails
        """
        self._auth.do_post(
            MgmtV1.access_key_update_path,
            {"id": id, "name": name, "description": description},
            pswd=self._auth.management_key,
        )

    def deactivate(
        self,
        id: str,
    ):
        """
        Deactivate an existing access key. IMPORTANT: This deactivated key will not be usable from this stage.
        It will, however, persist, and can be activated again if needed.

        Args:
        id (str): The id of the access key to be deactivated.

        Raise:
        AuthException: raised if deactivation operation fails
        """
        self._auth.do_post(
            MgmtV1.access_key_deactivate_path,
            {"id": id},
            pswd=self._auth.management_key,
        )

    def activate(
        self,
        id: str,
    ):
        """
        Activate an existing access key. IMPORTANT: Only deactivated keys can be activated again,
        and become usable once more. New access keys are active by default.

        Args:
        id (str): The id of the access key to be activate.

        Raise:
        AuthException: raised if activation operation fails
        """
        self._auth.do_post(
            MgmtV1.access_key_activate_path,
            {"id": id},
            pswd=self._auth.management_key,
        )

    def delete(
        self,
        id: str,
    ):
        """
        Delete an existing access key. IMPORTANT: This action is irreversible. Use carefully.

        Args:
        id (str): The id of the access key to be deleted.

        Raise:
        AuthException: raised if creation operation fails
        """
        self._auth.do_post(
            MgmtV1.access_key_delete_path,
            {"id": id},
            pswd=self._auth.management_key,
        )

    @staticmethod
    def _compose_create_body(
        name: str,
        expire_time: int,
        role_names: List[str],
        key_tenants: List[AssociatedTenant],
        user_id: Optional[str] = None,
        custom_claims: Optional[dict] = None,
        description: Optional[str] = None,
        permitted_ips: Optional[List[str]] = None,
    ) -> dict:
        return {
            "name": name,
            "expireTime": expire_time,
            "roleNames": role_names,
            "keyTenants": associated_tenants_to_dict(key_tenants),
            "userId": user_id,
            "customClaims": custom_claims,
            "description": description,
            "permittedIps": permitted_ips,
        }



from typing import Any, List, Optional

from descope._auth_base import AuthBase
from descope.management.common import MgmtV1


class Tenant(AuthBase):
    def create(
        self,
        name: str,
        id: Optional[str] = None,
        self_provisioning_domains: Optional[List[str]] = None,
        custom_attributes: Optional[dict] = None,
        enforce_sso: Optional[bool] = False,
        disabled: Optional[bool] = False,
    ) -> dict:
        """
        Create a new tenant with the given name. Tenant IDs are provisioned automatically, but can be provided
        explicitly if needed. Both the name and ID must be unique per project.

        Args:
        name (str): The tenant's name
        id (str): Optional tenant ID.
        self_provisioning_domains (List[str]): An optional list of domain that are associated with this tenant.
            Users authenticating from these domains will be associated with this tenant.
        custom_attributes (dict): Optional, set the different custom attributes values of the keys that were previously configured in Descope console app
        enforce_sso (bool): Optional, login to the tenant is possible only using the configured sso
        disabled (bool): Optional, login to the tenant will be disabled 

        Return value (dict):
        Return dict in the format
             {"id": <id>}

        Raise:
        AuthException: raised if creation operation fails
        """
        self_provisioning_domains = (
            [] if self_provisioning_domains is None else self_provisioning_domains
        )

        uri = MgmtV1.tenant_create_path
        response = self._auth.do_post(
            uri,
            Tenant._compose_create_update_body(
                name, id, self_provisioning_domains, custom_attributes, enforce_sso, disabled
            ),
            pswd=self._auth.management_key,
        )
        return response.json()

    def update(
        self,
        id: str,
        name: str,
        self_provisioning_domains: Optional[List[str]] = None,
        custom_attributes: Optional[dict] = None,
        enforce_sso: Optional[bool] = False,
        disabled: Optional[bool] = False,
    ):
        """
        Update an existing tenant with the given name and domains. IMPORTANT: All parameters are used as overrides
        to the existing tenant. Empty fields will override populated fields. Use carefully.

        Args:
        id (str): The ID of the tenant to update.
        name (str): Updated tenant name
        self_provisioning_domains (List[str]): An optional list of domain that are associated with this tenant.
            Users authenticating from these domains will be associated with this tenant.
        custom_attributes (dict): Optional, set the different custom attributes values of the keys that were previously configured in Descope console app
        enforce_sso (bool): Optional, login to the tenant is possible only using the configured sso
        disabled (bool): Optional, login to the tenant will be disabled 

        Raise:
        AuthException: raised if creation operation fails
        """
        self_provisioning_domains = (
            [] if self_provisioning_domains is None else self_provisioning_domains
        )

        uri = MgmtV1.tenant_update_path
        self._auth.do_post(
            uri,
            Tenant._compose_create_update_body(
                name, id, self_provisioning_domains, custom_attributes, enforce_sso, disabled
            ),
            pswd=self._auth.management_key,
        )

    def delete(
        self,
        id: str,
        cascade: bool = False,
    ):
        """
        Delete an existing tenant. IMPORTANT: This action is irreversible. Use carefully.

        Args:
        id (str): The ID of the tenant that's to be deleted.

        Raise:
        AuthException: raised if creation operation fails
        """
        uri = MgmtV1.tenant_delete_path
        self._auth.do_post(
            uri, {"id": id, "cascade": cascade}, pswd=self._auth.management_key
        )

    def load(
        self,
        id: str,
    ) -> dict:
        """
        Load tenant by id.

        Args:
        id (str): The ID of the tenant to load.

        Return value (dict):
        Return dict in the format
             {"id": <id>, "name": <name>, "selfProvisioningDomains": [], "customAttributes: {}, "createdTime": <timestamp>}
        Containing the loaded tenant information.

        Raise:
        AuthException: raised if load operation fails
        """
        response = self._auth.do_get(
            uri=MgmtV1.tenant_load_path,
            params={"id": id},
            pswd=self._auth.management_key,
        )
        return response.json()

    def load_all(
        self,
    ) -> dict:
        """
        Load all tenants.

        Return value (dict):
        Return dict in the format
             {"tenants": [{"id": <id>, "name": <name>, "selfProvisioningDomains": [], customAttributes: {}, "createdTime": <timestamp>}]}
        Containing the loaded tenant information.

        Raise:
        AuthException: raised if load operation fails
        """
        response = self._auth.do_get(
            uri=MgmtV1.tenant_load_all_path,
            pswd=self._auth.management_key,
        )
        return response.json()

    def search_all(
        self,
        ids: Optional[List[str]] = None,
        names: Optional[List[str]] = None,
        self_provisioning_domains: Optional[List[str]] = None,
        custom_attributes: Optional[dict] = None,
    ) -> dict:
        """
        Search all tenants.

        Args:
        ids (List[str]): Optional list of tenant IDs to filter by
        names (List[str]): Optional list of names to filter by
        self_provisioning_domains (List[str]): Optional list of self provisioning domains to filter by
        custom_attributes (dict): Optional search for a attribute with a given value

        Return value (dict):
        Return dict in the format
             {"tenants": [{"id": <id>, "name": <name>, "selfProvisioningDomains": [], customAttributes:{}}]}
        Containing the loaded tenant information.

        Raise:
        AuthException: raised if load operation fails
        """
        response = self._auth.do_post(
            uri=MgmtV1.tenant_search_all_path,
            body={
                "tenantIds": ids,
                "tenantNames": names,
                "tenantSelfProvisioningDomains": self_provisioning_domains,
                "customAttributes": custom_attributes,
            },
            pswd=self._auth.management_key,
        )
        return response.json()

    @staticmethod
    def _compose_create_update_body(
        name: str,
        id: Optional[str],
        self_provisioning_domains: List[str],
        custom_attributes: Optional[dict] = None,
        enforce_sso: Optional[bool] = False,
        disabled: Optional[bool] = False,
    ) -> dict:
        body: dict[str, Any] = {
            "name": name,
            "id": id,
            "selfProvisioningDomains": self_provisioning_domains,
            "enforceSSO": enforce_sso,
            "disabled": disabled
        }
        if custom_attributes is not None:
            body["customAttributes"] = custom_attributes
        return body



from typing import List, Optional

from descope._auth_base import AuthBase
from descope.management.common import MgmtV1


class Group(AuthBase):
    def load_all_groups(
        self,
        tenant_id: str,
    ) -> dict:
        """
        Load all groups for a specific tenant id.

        Args:
        tenant_id (str): Tenant ID to load groups from.

        Return value (dict):
        Return dict in the format
             [
                {
                    "id": <group id>,
                    "display": <display name>,
                    "members":[
                        {
                            "loginId": <loginId>,
                            "userId": <userId>,
                            "display": <display name>
                        }
                    ]
                }
            ]
        Containing the loaded groups information.

        Raise:
        AuthException: raised if load operation fails
        """
        response = self._auth.do_post(
            MgmtV1.group_load_all_path,
            {
                "tenantId": tenant_id,
            },
            pswd=self._auth.management_key,
        )
        return response.json()

    def load_all_groups_for_members(
        self,
        tenant_id: str,
        user_ids: Optional[List[str]] = None,
        login_ids: Optional[List[str]] = None,
    ) -> dict:
        """
        Load all groups for the provided user IDs or login IDs.

        Args:
        tenant_id (str): Tenant ID to load groups from.
        user_ids (List[str]): Optional List of user IDs, with the format of "U2J5ES9S8TkvCgOvcrkpzUgVTEBM" (example), which can be found on the user's JWT.
        login_ids (List[str]): Optional List of login IDs, how the users identify when logging in.

        Return value (dict):
        Return dict in the format
             [
                {
                    "id": <group id>,
                    "display": <display name>,
                    "members":[
                        {
                            "loginId": <loginId>,
                            "userId": <userId>,
                            "display": <display name>
                        }
                    ]
                }
            ]
        Containing the loaded groups information.

        Raise:
        AuthException: raised if load operation fails
        """
        user_ids = [] if user_ids is None else user_ids
        login_ids = [] if login_ids is None else login_ids

        response = self._auth.do_post(
            MgmtV1.group_load_all_for_member_path,
            {
                "tenantId": tenant_id,
                "loginIds": login_ids,
                "userIds": user_ids,
            },
            pswd=self._auth.management_key,
        )
        return response.json()

    def load_all_group_members(
        self,
        tenant_id: str,
        group_id: str,
    ) -> dict:
        """
        Load all members of the provided group id.

        Args:
        tenant_id (str): Tenant ID to load groups from.
        group_id (str): Group ID to load members for.

        Return value (dict):
        Return dict in the format
             [
                {
                    "id": <group id>,
                    "display": <display name>,
                    "members":[
                        {
                            "loginId": <loginId>,
                            "userId": <userId>,
                            "display": <display name>
                        }
                    ]
                }
            ]
        Containing the loaded groups information.

        Raise:
        AuthException: raised if load operation fails
        """
        response = self._auth.do_post(
            MgmtV1.group_load_all_group_members_path,
            {
                "tenantId": tenant_id,
                "groupId": group_id,
            },
            pswd=self._auth.management_key,
        )
        return response.json()



from typing import Optional

from descope._auth_base import AuthBase
from descope.common import EndpointsV1, LoginOptions, validate_refresh_token_provided
from descope.exceptions import ERROR_TYPE_INVALID_ARGUMENT, AuthException


class SSO(AuthBase):
    def start(
        self,
        tenant: str,
        return_url: Optional[str] = None,
        login_options: Optional[LoginOptions] = None,
        refresh_token: Optional[str] = None,
        prompt: Optional[str] = None,
        sso_id: Optional[str] = None,
    ) -> dict:
        """
        Start tenant sso session (saml/oidc based on tenant settings)

        Return value (dict): the redirect url for the login page
        Return dict in the format
             {'url': 'http://dummy.com/login..'}
        """
        if not tenant:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "Tenant cannot be empty"
            )

        validate_refresh_token_provided(login_options, refresh_token)

        uri = EndpointsV1.auth_sso_start_path
        params = SSO._compose_start_params(
            tenant,
            return_url if return_url else "",
            prompt if prompt else "",
            sso_id if sso_id else "",
        )
        response = self._auth.do_post(
            uri, login_options.__dict__ if login_options else {}, params, refresh_token
        )

        return response.json()

    def exchange_token(self, code: str) -> dict:
        uri = EndpointsV1.sso_exchange_token_path
        return self._auth.exchange_token(uri, code)

    @staticmethod
    def _compose_start_params(
        tenant: str, return_url: str, prompt: str, sso_id: str
    ) -> dict:
        res = {"tenant": tenant}
        if return_url is not None and return_url != "":
            res["redirectURL"] = return_url
        if prompt is not None and prompt != "":
            res["prompt"] = prompt
        if sso_id is not None and sso_id != "":
            res["ssoId"] = sso_id
        return res



from typing import Iterable, Optional, Union

from descope._auth_base import AuthBase
from descope.common import (
    REFRESH_SESSION_COOKIE_NAME,
    EndpointsV1,
    LoginOptions,
    validate_refresh_token_provided,
)
from descope.exceptions import ERROR_TYPE_INVALID_ARGUMENT, AuthException


class TOTP(AuthBase):
    def sign_up(self, login_id: str, user: Optional[dict] = None) -> dict:
        """
        Sign up (create) a new user using their email or phone number.
            (optional) Include additional user metadata that you wish to save.

        Args:
        login_id (str): The login ID of the user being validated
        user (dict) optional: Preserve additional user metadata in the form of,
             {"name": "Desmond Copeland", "phone": "2125551212", "email": "des@cope.com"}

        Return value (dict):
        Return dict in the format
             {"provisioningURL": "", "image": "", "key": ""}
        Includes 3 different ways to allow the user to save their credentials in
        their authenticator app, either by clicking the provisioning URL, scanning the QR
        image or inserting the key manually.

        Raise:
        AuthException: raised if sign-up operation fails
        """

        if not login_id:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "Identifier cannot be empty"
            )

        uri = EndpointsV1.sign_up_auth_totp_path
        body = TOTP._compose_signup_body(login_id, user)
        response = self._auth.do_post(uri, body)

        return response.json()

    def sign_in_code(
        self,
        login_id: str,
        code: str,
        login_options: Optional[LoginOptions] = None,
        refresh_token: Optional[str] = None,
        audience: Union[str, None, Iterable[str]] = None,
    ) -> dict:
        """
        Sign in by verifying the validity of a TOTP code entered by an end user.

        Args:
        login_id (str): The login ID of the user being validated
        code (str): The authenticator app code provided by the end user
        login_options (LoginOptions): Optional advanced controls over login parameters
        refresh_token: Optional refresh token is needed for specific login options

        Return value (dict):
        Return dict in the format
             {"jwts": [], "user": "", "firstSeen": "", "error": ""}
        Includes all the jwts tokens (session token, refresh token), token claims, and user information

        Raise:
        AuthException: raised if the TOTP code is not valid or if code verification failed
        """

        if not login_id:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "Identifier cannot be empty"
            )

        if not code:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "Code cannot be empty"
            )

        validate_refresh_token_provided(login_options, refresh_token)

        uri = EndpointsV1.verify_totp_path
        body = TOTP._compose_signin_body(login_id, code, login_options)
        response = self._auth.do_post(uri, body, None, refresh_token)

        resp = response.json()
        jwt_response = self._auth.generate_jwt_response(
            resp, response.cookies.get(REFRESH_SESSION_COOKIE_NAME, None), audience
        )
        return jwt_response

    def update_user(self, login_id: str, refresh_token: str) -> None:
        """
        Add TOTP to an existing logged in user using their refresh token.

        Args:
        login_id (str): The login ID of the user whose information is being updated
        refresh_token (str): The session's refresh token (used for verification)

        Return value (dict):
        Return dict in the format
             {"provisioningURL": "", "image": "", "key": ""}
        Includes 3 different ways to allow the user to save their credentials in
        their authenticator app, either by clicking the provisioning URL, scanning the QR
        image or inserting the key manually.

        Raise:
        AuthException: raised if refresh token is invalid or update operation fails
        """

        if not login_id:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "Identifier cannot be empty"
            )

        if not refresh_token:
            raise AuthException(
                400, ERROR_TYPE_INVALID_ARGUMENT, "Refresh token cannot be empty"
            )

        uri = EndpointsV1.update_totp_path
        body = TOTP._compose_update_user_body(login_id)
        response = self._auth.do_post(uri, body, None, refresh_token)

        return response.json()

    @staticmethod
    def _compose_signup_body(login_id: str, user: Optional[dict]) -> dict:
        body: dict[str, str | dict] = {"loginId": login_id}
        if user is not None:
            body["user"] = user
        return body

    @staticmethod
    def _compose_signin_body(
        login_id: str, code: str, login_options: Optional[LoginOptions] = None
    ) -> dict:
        return {
            "loginId": login_id,
            "code": code,
            "loginOptions": login_options.__dict__ if login_options else {},
        }

    @staticmethod
    def _compose_update_user_body(login_id: str) -> dict:
        return {"loginId": login_id}
